/*
 * ICFP 2002 Programming Contest
 * Entry by Seth Bruder
 *
 */

package sethbruder.icfp2002.entry.player;

import java.util.HashSet;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

import sethbruder.icfp2002.entry.*;

final class BaseState {

	private static int capacity = -1;

	public final boolean isPermanent;

	public final int x;
	public final int y;

	// Set of PackageInfo
	public final Set packages;
	
	public int timeOfLastVisit;

	private BaseState(boolean isPermanent, int x, int y, int timeOfLastVisit) {
		this.isPermanent = isPermanent;
		this.x = x;
		this.y = y;
		this.packages = new HashSet();
		this.timeOfLastVisit = timeOfLastVisit;
	}

	public static Map getInitBaseStates(PlayerState ps, InitInfo ii) {

		capacity = ii.capacity;

		boolean foundPermBase = false;
		final byte[][] board = ii.board.board;

		Map rv = new HashMap();
		for (int x = 0; x < board.length; x++)
			for (int y = 0; y < board[0].length; y++) {
				if (board[x][y] == Board.BASE) {
					rv.put(new Board.Point(x, y),
						   new BaseState(true, x, y, 0));
					foundPermBase = true;
				}
			}

		if (!foundPermBase)
			throw new RuntimeException("No home bases on board.");

		BaseState.update(rv, ps.x, ps.y, ii.turnOutcome.packagesHere, 0);

		return rv;
	}

	public static void update(Map baseMap, int x, int y, Set packages,
							  int time) {

		if (capacity < 0) throw new RuntimeException();

		packages = removeOverCapacityPackages(packages);

		Board.Point p = new Board.Point(x, y);
		BaseState bs = (BaseState)baseMap.get(p);
		if (bs == null) {
			if (!packages.isEmpty()) {
				bs = new BaseState(false, x, y, time);
				bs.packages.addAll(packages);
				baseMap.put(p, bs);
			}
		}
		else {
			if (packages.isEmpty()) {
				if (!bs.isPermanent) baseMap.remove(p);
				else bs.timeOfLastVisit = time;
			}
			else {
				bs.packages.clear();
				bs.packages.addAll(packages);
				bs.timeOfLastVisit = time;
			}
		}
	}

	/**
	 * (x, y) is current position of player (after turn)
	 * @return score increment
	 **/
	public static int 
		update(Map baseMap, int x, int y, PlayerMoveOutcome pmo) {

		if (capacity < 0) throw new RuntimeException();

		boolean sawMove = false;
		Move m;
		int score = 0;

		// keep track of ids because one item can be both picked up
		// and dropped in a single turn; only the last one counts
		Set idsSeen = new HashSet();
		Integer idSeen;

		for (int i = (pmo.moves.length - 1); i >= 0; i--) {
			m = pmo.moves[i]; 

			switch (m.type) { // backtrack
				case Move.NORTH:
					y--;
					sawMove = true;
					break;
				case Move.SOUTH:
					y++;
					sawMove = true;
					break;
				case Move.EAST: 
					x--;
					sawMove = true;
					break;
				case Move.WEST: 
					x++;
					sawMove = true;
					break;
				case Move.PICK: 
					for (int j = 0; j < m.ids.length; j++) {
						idSeen = new Integer(m.ids[j]);
						if (!idsSeen.contains(idSeen)) {
							pick(baseMap, x, y, m.ids[j]);
							idsSeen.add(idSeen);
						}
					}
					break;
				case Move.DROP: 
					for (int j = 0; j < m.ids.length; j++) {
						idSeen = new Integer(m.ids[j]);
						if (!idsSeen.contains(idSeen)) {
							score += drop(baseMap, x, y, m.ids[j], sawMove);
							idsSeen.add(idSeen);
						}
					}
					break;
				case Move.POSITION:
					x = m.x;
					y = m.y;
					break;
				default:
					throw new RuntimeException();
			}
		}

		return score;
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();

		sb.append("perm=" + isPermanent + ", x=" + x +", y=" + y
				  + ", time=" + timeOfLastVisit + ", pkgs=[");
		PackageInfo[] pis
			= (PackageInfo[])packages.toArray(new PackageInfo[0]);
		for (int i = 0; i < pis.length; i++) {
			if (i > 0) sb.append(", ");
			sb.append(pis[i].toString());
		}
		sb.append("]");

		return sb.toString();
	}

	private static void pick(Map baseMap, int x, int y, int id) {
		Board.Point p = new Board.Point(x, y);
		BaseState bs = (BaseState)baseMap.get(p);
		if (bs == null) return;

		bs.packages.remove(PackageInfo.get(id));
		if (bs.packages.isEmpty() && !bs.isPermanent) baseMap.remove(p);
	}

	private static int drop(Map baseMap, int x, int y, int id,
							boolean sawMove) {

		PackageInfo pi = PackageInfo.get(id);

		// if drop not caused by a push, and we don't know the dest, return
		if ((pi.x == -1) && !sawMove) return 0;

		if ((x == pi.x) && (y == pi.y)) return pi.weight;

		if (pi.weight > capacity) return 0;

		Board.Point p = new Board.Point(x, y);
		BaseState bs = (BaseState)baseMap.get(p);
		if (bs == null) {
			bs = new BaseState(false, x, y, 0);
			baseMap.put(p, bs);
		}
		bs.packages.add(pi);

		return 0;
	}

	private static Set removeOverCapacityPackages(Set packages) {

		Set filteredPackages = new HashSet();
		Iterator it = packages.iterator();
		while (it.hasNext()) {
			PackageInfo pi = (PackageInfo)it.next();
			if (pi.weight <= capacity) filteredPackages.add(pi);
		}
		return filteredPackages;
	}
}
